unit IntrThread;

interface
uses windows, classes, SysUtils;

{.$DEFINE DEBUG}
{$IFDEF DEBUG}
{.$DEFINE DEBUG2}
{$ENDIF}
const
 MaxTicsCount = 10000;
 cDivMeasuringPeriod = 100;
 cStartMeasuringPeriod = 10 * cDivMeasuringPeriod;
 cStartReabilitationPeriod = 5 * cStartMeasuringPeriod;
 cStartPoorPeriod = 30 * cStartMeasuringPeriod;
 cStartBadPeriod  = 100 * cStartMeasuringPeriod;
 cStartRoughPeriod = 600 * cStartMeasuringPeriod;

 cMinReabilitationLevel = 0.5;
 cMaxReabilitationLevel = 0.9;
 cMinBadLevel = 0.7;
 cMaxBadLevel = 0.999;

  {
  TDateTime
  WaitForSingleObjectEx
  }
type
  TThreadHistoryRec = record
    CreationTime,	// when the thread was created
    KernelTime,	// time the thread has spent in kernel mode
    UserTime  : TFileTime; // time the thread has spent in user mode
    KernelUsage, UserUsage : smallint;
  end;
    {
     GetThreadTimes
     FileTimeToSystemTime
    }
  TAfterRoughMode = (armIgnore, armTerminateThread, armTerminateProcess,
  armException);

  TMainThreadState = (mtsUndef, mtsGood, mtsPoor, mtsBad, mtsRough);

  TIntruderThread = class(TThread)
   private
     fAfterRoughMode : TAfterRoughMode;
     fAccessCriticalSection : TRTLCriticalSection;
     fHistoryKBuff : array[0..MaxTicsCount-1] of TFileTime;
     fHistoryTBuff : array[0..MaxTicsCount-1] of TFileTime;
     fMeasuringPeriod,
     fReabilitationPeriod,
     fPoorPeriod,
     fBadPeriod,
     fRoughPeriod : cardinal;
     fCurrentState : TMainThreadState;
     fBufferStartPosition, fBufferEndPosition : integer;
     fStartingKernelValue : TFileTime;
     fStartingKernelValueMSec : cardinal;
     fWasLoop : boolean;
     fLastStageIndex : integer;

     fReabilitationLevel, fBadLevel, fAverageCPUusageMax  : double;
     fPoorPeriodCount,
     fBadPeriodCount,
     fRoughPeriodCount : cardinal;

     function GetMinMeasuringPeriod : cardinal;
     function GetMinReabilitationPeriod  : cardinal;
     function GetMinPoorPeriod  : cardinal;
     function GetMinBadPeriod   : cardinal;
     function GetMinRoughPeriod  : cardinal;

     function GetMaxMeasuringPeriod  : cardinal;
     function GetMaxReabilitationPeriod  : cardinal;
     function GetMaxPoorPeriod  : cardinal;
     function GetMaxBadPeriod   : cardinal;
     function GetMaxRoughPeriod : cardinal;

     procedure SetReabilitationLevel (aValue : double);
     procedure SetBadLevel (aValue : double);
     procedure SetMeasuringPeriod (aValue : cardinal);
     procedure SetReabilitationPeriod (aValue : cardinal);
     procedure SetPoorPeriod (aValue : cardinal);
     procedure SetBadPeriod (aValue : cardinal);
     procedure SetRoughPeriod (aValue : cardinal);
     function  GetIsNoRecords : boolean;
     function  GetKernelTime (var aSystemTime : TFileTime) : TFileTime;
     function  GetHistoryKBuff (Index : integer) : TFileTime;
     function  GetHistoryTBuff (Index : integer) : TFileTime;
     function  GetAverageCPUusage : double;
     function  DecPriority : boolean;
     function  IncPriority : boolean;
     procedure SetCurrentState (aValue : TMainThreadState);
   protected
     procedure Execute; override;
     procedure SetHandlesOfMain;
     procedure FreeHandlesOfMain;
     procedure SetHistoryPointNow;
     property  HistoryKBuff[Index : integer] : TFileTime read GetHistoryKBuff;
     property  HistoryTBuff[Index : integer] : TFileTime read GetHistoryTBuff;
   public
     MainThreadHandle : THandle;
     MainThreadID : THandle;
     LastThreadHistoryRec  : TThreadHistoryRec;
     property  AfterRoughMode : TAfterRoughMode read fAfterRoughMode
                                        write fAfterRoughMode;
     property  LastStageIndex : integer read fLastStageIndex
                                        write fLastStageIndex;
     property  ReabilitationLevel : double
                read fReabilitationLevel write SetReabilitationLevel;
     property  BadLevel : double
                read fBadLevel write SetBadLevel;
     property  IsNoRecords : boolean read GetIsNoRecords;
     property  CurrentState : TMainThreadState read fCurrentState
                                               write SetCurrentState;
     property  MeasuringPeriod : cardinal
                read fMeasuringPeriod write SetMeasuringPeriod;
     property  ReabilitationPeriod : cardinal
                read fReabilitationPeriod write SetReabilitationPeriod;
     property  PoorPeriod : cardinal
                read fPoorPeriod write SetPoorPeriod;
     property  BadPeriod : cardinal
                read fBadPeriod write SetBadPeriod;
     property  RoughPeriod : cardinal
                read fRoughPeriod write SetRoughPeriod;

     property  PoorPeriodCount : cardinal read fPoorPeriodCount;
     property  BadPeriodCount : cardinal read fBadPeriodCount;
     property  RoughPeriodCount : cardinal read fRoughPeriodCount;

     property  AverageCPUusage : double read GetAverageCPUusage;
     property  AverageCPUusageMax : double read fAverageCPUusageMax;

     function  GetStatePeriod (aState : TMainThreadState) : cardinal;
     function  GetKernelTimeForLast (aMeasuringPeriod : cardinal;
                   var Undef : boolean) : double;
     procedure LockAccess;
     procedure UnLockAccess;
     constructor Create(CreateSuspended: Boolean); virtual;
     destructor Destroy; override;
  end;

procedure ExecuteDummyCalk ( i, j : integer; var  R : Extended);
{
function DateTimeToMilisec (DT : TDateTime) : cardinal;
}
function FileTimeToMilisec (FT : TFileTime) : cardinal;
function CompareFileTimeToMilisec (FT1, FT2 : TFileTime) : cardinal;
function FileTimeSub (FT1, FT2 : TFileTime) : TFileTime;
function FileTimeAdd (FT1, FT2 : TFileTime) : TFileTime;

implementation

{$IFDEF DEBUG}
var fZ : TextFile;
procedure LogResults(IT : TIntruderThread);
begin
 with IT do
  try
   Writeln (fz, 'BufferStartPosition=',fBufferStartPosition,
                '| BufferEndPosition=',fBufferEndPosition,
         '| HistoryPoint.Low=', HistoryKBuff[fBufferEndPosition].dwLowDateTime,
         '| HistoryPoint.Hi=', HistoryKBuff[fBufferEndPosition].dwHighDateTime);
   Flush(fz);
  except
  end;
end;

procedure LogParams(St : string);
begin
  try
   Writeln (fz, St);
   Flush(fz);
  except
  end;
end;
{$ENDIF}

function FileTimeSub (FT1, FT2 : TFileTime) : TFileTime;
var tmp1, tmp2 : DWORD;
begin
 asm
   MOV EAX, FT1.dwLowDateTime
   SUB EAX, FT2.dwLowDateTime
   MOV tmp1, EAX
   MOV EAX, FT1.dwHighDateTime
   SBB EAX, FT2.dwHighDateTime
   MOV tmp2, EAX
 end;
 Result.dwLowDateTime := tmp1;
 Result.dwHighDateTime := tmp2;
end;

function FileTimeAdd (FT1, FT2 : TFileTime) : TFileTime;
var tmp1, tmp2 : DWORD;
begin
 asm
   MOV EAX, FT1.dwLowDateTime
   ADD EAX, FT2.dwLowDateTime
   MOV tmp1, EAX
   MOV EAX, FT1.dwHighDateTime
   ADC EAX, FT2.dwHighDateTime
   MOV tmp2, EAX
 end;
 Result.dwLowDateTime := tmp1;
 Result.dwHighDateTime := tmp2;
end;

function CompareFileTimeToMilisec (FT1, FT2 : TFileTime) : cardinal;
begin
  Result := FileTimeToMilisec(FileTimeSub(FT1, FT2));
end;
(*
function DateTimeToMilisec (DT : TDateTime) : cardinal;
var TS : TTimeStamp;
begin
 TS := DateTimeToTimeStamp (DT);
 {
 Result := TS.Time;
 }
 Result := Round(DT*24*3600*1000);
end;
*)
function FileTimeToMilisec (FT : TFileTime) : cardinal;
var
    ST : TSystemTime;
begin
 FileTimeToSystemTime(FT, ST);


 Result :=
  ST.wMilliseconds + ST.wSecond*1000 + ST.wMinute*60000 +
   ST.wHour*60000*60 + (ST.wDay-1)*60000*60*24;
end;

function TIntruderThread.GetAverageCPUusage : double;
var TmpCreationTime, TmpExitTime, TmpKernelTime, TmpUserTime,
    aSystemTime : TFileTime;
    afSystemTime : TSystemTime;
    CompFT, Comp2FT : cardinal;
begin
 Result := 0;
 GetSystemTime(afSystemTime);
 GetThreadTimes (MainThreadHandle,
     TmpCreationTime, TmpExitTime, TmpKernelTime, TmpUserTime);
 TmpKernelTime := FileTimeAdd (TmpKernelTime, TmpUserTime);
 SystemTimeToFileTime (afSystemTime, aSystemTime);

  CompFT := FileTimeToMilisec (TmpKernelTime);

  Comp2FT := CompareFileTimeToMilisec(aSystemTime, TmpCreationTime);
  if not(Comp2FT = 0) then
     Result := CompFT  / Comp2FT;
  if Result > fAverageCPUusageMax then
    fAverageCPUusageMax := Result;
 {$IFDEF DEBUG2}

 try
  Writeln(fz, 'GetAverageCPUusage: '+
   'TmpKernelTime.h = ', TmpKernelTime.dwHighDateTime,
   '; TmpKernelTime.l = ', TmpKernelTime.dwLowDateTime,
   '; TmpCreationTime.h = ', TmpCreationTime.dwHighDateTime,
   '; TmpCreationTime.l = ', TmpCreationTime.dwLowDateTime,
   '; aSystemTime.h = ', aSystemTime.dwHighDateTime,
   '; aSystemTime.l = ', aSystemTime.dwLowDateTime,
   '; CompFT = ',CompFT, '; Comp2FT = ',Comp2FT,
   '; Res = ',Result);
  Flush(fz);
 finally
 end;

 {$ENDIF}

end;

constructor TIntruderThread.Create(CreateSuspended: Boolean);
var i : integer;
begin
 inherited Create(True);
 fPoorPeriodCount := 0;
 fBadPeriodCount := 0;
 fRoughPeriodCount := 0;
 fAfterRoughMode := armIgnore;
 fMeasuringPeriod := cStartMeasuringPeriod;
 fReabilitationPeriod := cStartReabilitationPeriod;
 fPoorPeriod := cStartPoorPeriod;
 fBadPeriod  := cStartBadPeriod;
 fRoughPeriod := cStartRoughPeriod;
 fCurrentState := mtsUndef;
 fBufferStartPosition := 0;
 fBufferEndPosition := -1;
 fLastStageIndex := 0;
 fWasLoop := False;
 fReabilitationLevel := cMinReabilitationLevel;
 fBadLevel := cMaxBadLevel;
 fAverageCPUusageMax := 0;
 FreeOnTerminate := True;
 for i:=0 to MaxTicsCount-1 do
  begin
   fHistoryKBuff[i].dwLowDateTime := 0;
   fHistoryKBuff[i].dwHighDateTime := 0;
   fHistoryTBuff[i].dwLowDateTime := 0;
   fHistoryTBuff[i].dwHighDateTime := 0;
  end;
 SetHandlesOfMain;
end;

function TIntruderThread.GetMinMeasuringPeriod : cardinal;
begin
 Result := cDivMeasuringPeriod;
end;

function TIntruderThread.GetMinReabilitationPeriod  : cardinal;
begin
 Result := 2*fMeasuringPeriod;
end;

function TIntruderThread.GetMinPoorPeriod  : cardinal;
begin
 Result := fReabilitationPeriod+fMeasuringPeriod;
end;

function TIntruderThread.GetMinBadPeriod   : cardinal;
begin
 Result := fPoorPeriod+fMeasuringPeriod;
end;

function TIntruderThread.GetMinRoughPeriod  : cardinal;
begin
 Result := fBadPeriod+fMeasuringPeriod;
end;

function TIntruderThread.GetMaxMeasuringPeriod  : cardinal;
begin
 Result := 5000;
end;


function TIntruderThread.GetMaxReabilitationPeriod  : cardinal;
begin
 Result := (MaxTicsCount div 50) * fMeasuringPeriod;
end;


function TIntruderThread.GetMaxPoorPeriod  : cardinal;
begin
 Result := ((MaxTicsCount div 50)+1) * fMeasuringPeriod;
end;


function TIntruderThread.GetMaxBadPeriod   : cardinal;
begin
 Result := (MaxTicsCount div 10) * fMeasuringPeriod;
end;


function TIntruderThread.GetMaxRoughPeriod : cardinal;
begin
 Result := (MaxTicsCount - 2) * fMeasuringPeriod;
end;



function  TIntruderThread.GetStatePeriod (aState : TMainThreadState) : cardinal;
begin
 case aState of
   mtsGood : Result := fReabilitationPeriod;
   mtsPoor : Result := fPoorPeriod;
   mtsBad  : Result := fBadPeriod;
   mtsRough: Result := fRoughPeriod;
  else
   Result := fMeasuringPeriod;
 end;
end;

procedure TIntruderThread.SetCurrentState (aValue : TMainThreadState);
begin
 if aValue = fCurrentState then exit;
 {$IFDEF DEBUG}
 try
 Writeln(fz, 'SetCurrentState. aValue=', ord(aValue),
             '; fCurrentState=',ord(fCurrentState) );
 Flush(fz);
 finally
 end;
 {$ENDIF}
 if fCurrentState in [mtsGood, mtsPoor, mtsBad, mtsRough] then
  begin
   if aValue <> mtsGood then
     begin
      LastStageIndex := fBufferEndPosition;
      case aValue of
       mtsPoor  :  inc(fPoorPeriodCount);
       mtsBad   :  inc(fBadPeriodCount);
       mtsRough :  inc(fRoughPeriodCount);
      end;
     end;
   if fCurrentState > aValue then
      IncPriority
   else
      DecPriority;
  end;
 fCurrentState := aValue;
end;

function TIntruderThread.DecPriority : boolean;
var Pri, NewPri : integer;
    OK : boolean;
{    OK2 : BOOL;}
begin
 OK := False;
 Result := False;
 try
   Pri := GetThreadPriority(MainThreadHandle);
 {$IFDEF DEBUG}
 try
 Writeln(fz, '!!!!DecPriority. Pri=', Pri);
 Flush(fz);
 finally
 end;
 {$ENDIF}
   while not(OK) do
    begin
     case Pri of
        THREAD_PRIORITY_ABOVE_NORMAL :
         {Indicates 1 point above normal priority for the priority class.}
          NewPri := THREAD_PRIORITY_NORMAL;
        THREAD_PRIORITY_BELOW_NORMAL :
         {Indicates 1 point below normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_LOWEST;
        THREAD_PRIORITY_HIGHEST :
         {Indicates 2 points above normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_ABOVE_NORMAL;
        THREAD_PRIORITY_IDLE :
         {Indicates a base-priority level of 1 for IDLE_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS, or HIGH_PRIORITY_CLASS processes, and a base-priority level of 16 for REALTIME_PRIORITY_CLASS processes.}
           NewPri := Pri;
        THREAD_PRIORITY_LOWEST :
         {Indicates 2 points below normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_IDLE;
        THREAD_PRIORITY_NORMAL :
         {Indicates normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_BELOW_NORMAL;
        THREAD_PRIORITY_TIME_CRITICAL :
           NewPri := THREAD_PRIORITY_HIGHEST;
      else
        NewPri := Pri;
     end;
     if (NewPri = Pri) then
       exit;
     {OK2 := }SetThreadPriority(MainThreadHandle, NewPri);
     Pri := GetThreadPriority(MainThreadHandle);
     if Pri = NewPri then
      begin
        OK := True;
        Result := True;
        {$IFDEF DEBUG}
        try
        Writeln(fz, '!!!! DecPriority. OK! NewPri=', NewPri);
        Flush(fz);
        finally
        end;
        {$ENDIF}
      end
     else
       Pri := NewPri;
    end;
  except
  end;
end;

function TIntruderThread.IncPriority : boolean;
var Pri, NewPri : integer;
    OK : boolean;
{    OK2 : BOOL;}
begin
 OK := False;
 Result := False;
 try
   Pri := GetThreadPriority(MainThreadHandle);
   {$IFDEF DEBUG}
   try
   Writeln(fz, '!!!! IncPriority. Pri=', Pri);
   Flush(fz);
   finally
   end;
   {$ENDIF}
   while not(OK) do
    begin
     case Pri of
        THREAD_PRIORITY_ABOVE_NORMAL :
         {Indicates 1 point above normal priority for the priority class.}
          NewPri := THREAD_PRIORITY_HIGHEST;
        THREAD_PRIORITY_BELOW_NORMAL :
         {Indicates 1 point below normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_NORMAL;
        THREAD_PRIORITY_HIGHEST :
         {Indicates 2 points above normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_ABOVE_NORMAL;
        THREAD_PRIORITY_IDLE :
         {Indicates a base-priority level of 1 for IDLE_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS, or HIGH_PRIORITY_CLASS processes, and a base-priority level of 16 for REALTIME_PRIORITY_CLASS processes.}
           NewPri := THREAD_PRIORITY_LOWEST;
        THREAD_PRIORITY_LOWEST :
         {Indicates 2 points below normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_BELOW_NORMAL;
        THREAD_PRIORITY_NORMAL :
         {Indicates normal priority for the priority class.}
           NewPri := THREAD_PRIORITY_ABOVE_NORMAL;
        THREAD_PRIORITY_TIME_CRITICAL :
           NewPri := Pri;
      else
        NewPri := Pri;
     end;
     if (NewPri = Pri) then
       exit;
     {OK2 := }
     SetThreadPriority(MainThreadHandle, NewPri);
     Pri := GetThreadPriority(MainThreadHandle);
     if Pri = NewPri then
      begin
        OK := True;
        Result := True;
        {$IFDEF DEBUG}
        try
        Writeln(fz, '!!!! IncPriority. OK! NewPri=', NewPri);
        Flush(fz);
        finally
        end;
        {$ENDIF}
      end
     else
       Pri := NewPri;
    end;
  except
  end;
end;


function  TIntruderThread.GetHistoryKBuff (Index : integer) : TFileTime;
begin
 if (Index < 0) or (Index >= MaxTicsCount) then
   raise Exception.Create('Buffer index out of available range');
 Result := fHistoryKBuff[Index];
end;

function  TIntruderThread.GetHistoryTBuff (Index : integer) : TFileTime;
begin
 if (Index < 0) or (Index >= MaxTicsCount) then
   raise Exception.Create('Buffer index out of available range');
 Result := fHistoryTBuff[Index];
end;

procedure TIntruderThread.SetReabilitationLevel (aValue : double);
begin
 if aValue = fReabilitationLevel then
      exit;
 fReabilitationLevel := aValue;
 if fReabilitationLevel < cMinReabilitationLevel then
      fReabilitationLevel := cMinReabilitationLevel;
 if fReabilitationLevel > cMaxReabilitationLevel then
      fReabilitationLevel := cMaxReabilitationLevel;
 if fBadLevel <= fReabilitationLevel then
      fBadLevel := fReabilitationLevel+0.01;
end;

procedure TIntruderThread.SetBadLevel (aValue : double);
begin
 if aValue = fBadLevel then
   exit;
 fBadLevel := aValue;
 if fBadLevel < cMinBadLevel then
      fBadLevel := cMinBadLevel;
 if fBadLevel > cMaxBadLevel then
      fBadLevel := cMaxBadLevel;
 if fBadLevel <= fReabilitationLevel then
      fReabilitationLevel := fBadLevel-0.01;
end;


procedure TIntruderThread.SetMeasuringPeriod (aValue : cardinal);
begin
 if (aValue = fMeasuringPeriod) then
   exit;
 if (aValue < GetMinMeasuringPeriod) then
   fMeasuringPeriod := GetMinMeasuringPeriod
 else
  if (aValue > GetMaxMeasuringPeriod) then
    fMeasuringPeriod := GetMaxMeasuringPeriod
  else
    fMeasuringPeriod := (aValue div cDivMeasuringPeriod) * cDivMeasuringPeriod;
 if (aValue = fMeasuringPeriod) then
   exit;
 ReabilitationPeriod := fReabilitationPeriod;
end;

procedure TIntruderThread.SetReabilitationPeriod (aValue : cardinal);
begin
 fReabilitationPeriod := (aValue div fMeasuringPeriod) * fMeasuringPeriod;
 if (aValue = fReabilitationPeriod) then
   exit;
 if (aValue < GetMinReabilitationPeriod) then
   fReabilitationPeriod := GetMinReabilitationPeriod
 else
  if (aValue > GetMaxReabilitationPeriod) then
    fReabilitationPeriod := GetMaxReabilitationPeriod;
 PoorPeriod  := fPoorPeriod;
end;

procedure TIntruderThread.SetPoorPeriod (aValue : cardinal);
begin
 fPoorPeriod := (aValue div fMeasuringPeriod) * fMeasuringPeriod;
 if (aValue = fPoorPeriod) then
   exit;
 if (aValue < GetMinPoorPeriod) then
   fPoorPeriod := GetMinPoorPeriod
 else
  if (aValue > GetMaxPoorPeriod) then
    fPoorPeriod := GetMaxPoorPeriod;
 BadPeriod   := fBadPeriod;
end;

procedure TIntruderThread.SetBadPeriod (aValue : cardinal);
begin
 fBadPeriod := (aValue div fMeasuringPeriod) * fMeasuringPeriod;
 if (aValue = fBadPeriod) then
   exit;
 if (aValue < GetMinBadPeriod) then
   fBadPeriod := GetMinBadPeriod
 else
  if (aValue > GetMaxBadPeriod) then
    fBadPeriod := GetMaxBadPeriod;
 RoughPeriod := fRoughPeriod;
end;

procedure TIntruderThread.SetRoughPeriod (aValue : cardinal);
begin
 fRoughPeriod := (aValue div fMeasuringPeriod) * fMeasuringPeriod;
 if (aValue = fRoughPeriod) then
   exit;
 if (aValue < GetMinRoughPeriod) then
   fRoughPeriod := GetMinRoughPeriod
 else
  if (aValue > GetMaxRoughPeriod) then
    fRoughPeriod := GetMaxRoughPeriod;
end;

procedure TIntruderThread.SetHandlesOfMain;
begin
 InitializeCriticalSection(fAccessCriticalSection);
 DuplicateHandle (GetCurrentProcess, GetCurrentThread,
                  GetCurrentProcess, @MainThreadHandle,
                  0,
                  False, DUPLICATE_SAME_ACCESS);
 MainThreadID  := GetCurrentThreadID;
end;

procedure TIntruderThread.FreeHandlesOfMain;
begin
 DeleteCriticalSection(fAccessCriticalSection);
 CloseHandle(MainThreadHandle);
 MainThreadID := 0;
end;

procedure TIntruderThread.SetHistoryPointNow;
var ftNeedToChangeLast : boolean;
begin
 if IsNoRecords then
  begin
    fHistoryKBuff[0] := GetKernelTime (fHistoryTBuff[0]);
    fBufferEndPosition := 0;
    {$IFDEF DEBUG2} {CreateThread}
    LogResults(Self);
    {$ENDIF}
  end
 else
  begin
   ftNeedToChangeLast := (LastStageIndex = fBufferStartPosition);
   inc(fBufferEndPosition);
   if (fBufferEndPosition >= MaxTicsCount) then
     fBufferEndPosition := 0;
   fHistoryKBuff[fBufferEndPosition] :=
     GetKernelTime(fHistoryTBuff[fBufferEndPosition]);
   {$IFDEF DEBUG2}
    LogResults(Self);
   {$ENDIF}
   if not(fWasLoop) then
    begin
     if (fBufferEndPosition = 0) then
       begin
         fBufferStartPosition := 1;
         if ftNeedToChangeLast then
            LastStageIndex := fBufferStartPosition;
         fWasLoop := True;
       end
    end
   else
     begin
       fBufferStartPosition := fBufferEndPosition + 1;
       if (fBufferStartPosition >= MaxTicsCount) then
         fBufferStartPosition := 0;
       if ftNeedToChangeLast then
          LastStageIndex := fBufferStartPosition;
     end;
  end
end;

function  TIntruderThread.GetIsNoRecords : boolean;
begin
 Result := (fBufferEndPosition = -1);
end;

function  TIntruderThread.GetKernelTime
  (var aSystemTime : TFileTime) : TFileTime;
var TmpCreationTime, TmpExitTime, TmpKernelTime, TmpUserTime : TFileTime;
{$IFDEF DEBUG2}
    B : BOOL;
{$ENDIF}
    afSystemTime : TSystemTime;
begin
 Result.dwLowDateTime := 0;
 Result.dwHighDateTime := 0;
 GetSystemTime(afSystemTime);
 {$IFDEF DEBUG2}
 B :=
 {$ENDIF}
 GetThreadTimes (MainThreadHandle,
     TmpCreationTime, TmpExitTime, TmpKernelTime, TmpUserTime);
 TmpKernelTime := FileTimeAdd (TmpKernelTime, TmpUserTime);
 SystemTimeToFileTime (afSystemTime, aSystemTime);
 {$IFDEF DEBUG2}
    try
      Win32Check (B);
      LogParams('KernelTime.Hi = '+IntToHex(TmpKernelTime.dwHighDateTime, 8)+
            '; KernelTime.Lo = '+IntToHex(TmpKernelTime.dwLowDateTime, 8));
      LogParams('   SystemTime.Hi = '+IntToHex(aSystemTime.dwHighDateTime, 8)+
            '; SystemTime.Lo = '+IntToHex(aSystemTime.dwLowDateTime, 8))
    except
     on E:Exception do
      begin
        LogParams('KernelTime.Hi = '+IntToHex(TmpKernelTime.dwHighDateTime, 8)+
            '; KernelTime.Lo = '+IntToHex(TmpKernelTime.dwLowDateTime, 8)+
            '; RESULT = '+IntToHex(DWORD(B), 8)+E.Message);
        LogParams('   SystemTime.Hi = '+IntToHex(aSystemTime.dwHighDateTime, 8)+
            '; SystemTime.Lo = '+IntToHex(aSystemTime.dwLowDateTime, 8));
      end
    end;
 {$ENDIF}
 Result := TmpKernelTime;
 if IsNoRecords then
   begin
     fStartingKernelValue := TmpKernelTime;
     fStartingKernelValueMSec := FileTimeToMilisec(fStartingKernelValue);
   end;

end;

function TIntruderThread.GetKernelTimeForLast (aMeasuringPeriod : cardinal;
                                              var Undef : boolean) : double;
var tfStartIndex : integer;
    CompFT, Comp2FT : cardinal;
begin
 {$IFDEF DEBUG2}
 try
 Writeln(fz, 'GetKernelTimeForLast. MeasuringPeriod=', aMeasuringPeriod);
 Flush(fz);
 finally
 end;
 {$ENDIF}
  Result := 0;
  Undef := True;
  if (aMeasuringPeriod = 0) then Exit;
  tfStartIndex := fBufferEndPosition -
                       ABS(aMeasuringPeriod div fMeasuringPeriod);
  if tfStartIndex < 0 then
   begin
     if not(fWasLoop) then
       exit;

     tfStartIndex := MaxTicsCount - tfStartIndex;
     if tfStartIndex < 0 then Exit;
   end;
  if (tfStartIndex = fBufferEndPosition) then Exit;
  CompFT := CompareFileTimeToMilisec(
                 HistoryKBuff[fBufferEndPosition],
                 HistoryKBuff[tfStartIndex]
                 );
  Comp2FT := CompareFileTimeToMilisec(
                 HistoryTBuff[fBufferEndPosition],
                 HistoryTBuff[tfStartIndex]
                 );
  if Comp2FT = 0 then
    exit;
  Undef := False;
  Result := CompFT  / Comp2FT;
  if Result > fAverageCPUusageMax then
    fAverageCPUusageMax := Result;
 {$IFDEF DEBUG2}
 try
 Writeln(fz, 'GetKernelTimeForLast: BufferEndPosition = ',fBufferEndPosition,
 '; tfStartIndex=',tfStartIndex,'; CompFT = ',CompFT, '; Comp2FT = ',Comp2FT,
  '; Res = ',Result);
 Flush(fz);
 finally
 end;
 {$ENDIF}

end;


const a : array [0..9, 0..1] of Extended =
   (
    (1.15, 2.12),
    (11.45, 3.27),
    (9.7815, 2.29),
    (4.23, 2.20),
    (7.66, -2.321),
    (1.86, 8.62),
    (8.25, 9.12),
    (2.35, 3.12),
    (5.16, -1.12),
    (-4.13, 4.72)
    );

procedure ExecuteDummyCalk ( i, j : integer; var  R : Extended);
var ti : integer;
begin
 R := 0;
 for ti:= i to i+j do
  begin
   try
    R := R * SQRT(ABS(A[ti mod 10, 1])) - SQR(ABS(A[ti mod 10, 0]));
   except
   end
  end;
end;

procedure TIntruderThread.Execute;
var ResW : DWORD;
    Undef : boolean;
    ResD : double;
    tmpInt : integer;
    TmpState : TMainThreadState;
begin
 SetThreadPriority (GetCurrentThread, THREAD_PRIORITY_HIGHEST);
 LockAccess;
{$IFDEF DEBUG}
 try
 Writeln(fz, 'MeasuringPeriod = ',MeasuringPeriod);
 Writeln(fz, 'ReabilitationPeriod = ',ReabilitationPeriod);
 Writeln(fz, 'PoorPeriod = ',PoorPeriod);
 Writeln(fz, 'RoughPeriod = ',RoughPeriod);
 Flush(fz);
 finally
 end;
{$ENDIF}
 SetHistoryPointNow;
 UnLockAccess;
 while 1 = 1 do
  begin
   if Terminated then
       Exit;
   ResW := WaitForSingleObject (MainThreadHandle, fMeasuringPeriod);
   try
    LockAccess;
    case ResW of
     WAIT_OBJECT_0 :{MainThread Terminated}
              Exit;
     WAIT_FAILED : {Error waiting object}
              Exit;
     WAIT_TIMEOUT :
      begin
        SetHistoryPointNow;
        case CurrentState of
          mtsUndef :
             begin
               ResD := GetKernelTimeForLast (PoorPeriod, Undef);
               if not(Undef) then
                begin
                  CurrentState := mtsGood;
                  if (ResD >= BadLevel) then
                   begin
                    CurrentState := mtsPoor;
                    {LastStageIndex := fBufferEndPosition;}
                   end;
                end;
             end;
          mtsGood :
             begin
               ResD := GetKernelTimeForLast (PoorPeriod, Undef);
               if (ResD >= BadLevel) and not(Undef)then
                begin
                   CurrentState := mtsPoor;
                   {LastStageIndex := fBufferEndPosition;}
                end;
             end;
          else
             begin
               ResD := GetKernelTimeForLast (ReabilitationPeriod, Undef);
               if (ResD <= ReabilitationLevel)  and not(Undef) then
                 begin
                   TmpState := CurrentState;
                   TmpState := Pred(TmpState);
                   CurrentState := TmpState;
                   {LastStageIndex := fBufferEndPosition;}
                 end
               else
                begin
                  if fBufferEndPosition < LastStageIndex then
                     tmpInt := MaxTicsCount + fBufferEndPosition - LastStageIndex
                   else
                     tmpInt := fBufferEndPosition - LastStageIndex;
                       {$IFDEF DEBUG2}
                       try
                        Writeln(fz,'* fBufferEndPosition:',fBufferEndPosition,
                          '; LastStageIndex:',LastStageIndex,';  tmpInt:',tmpInt,
                          '; fBufferEndPosition:',fBufferEndPosition
                         );
                        Flush(fz);
                       finally
                       end;
                       {$ENDIF}
                  if ( (GetStatePeriod(CurrentState) div MeasuringPeriod)
                          <= tmpInt)  then
                    begin
                      ResD := GetKernelTimeForLast (
                                     GetStatePeriod(CurrentState), Undef
                                   );
                       {$IFDEF DEBUG2}
                       try
                        Writeln(fz,'** GetStatePeriod(CurrentState):',
                           GetStatePeriod(CurrentState),
                          '; Undef:', Undef, '; Result:',
                          ResD
                         );
                        Flush(fz);
                       finally
                       end;
                       {$ENDIF}
                      if (ResD >= BadLevel)  and not(Undef) then
                       begin
                       {$IFDEF DEBUG2}
                       try
                        Writeln(fz,' Decreasing level. CurrentState',
                         ord(CurrentState));
                        Flush(fz);
                       finally
                       end;
                       {$ENDIF}
                        if (CurrentState < High(CurrentState)) then
                         begin
                          TmpState := CurrentState;
                          TmpState := Succ(TmpState);
                          CurrentState := TmpState;
                         {LastStageIndex := fBufferEndPosition;}
                         end
                         else
                           case fAfterRoughMode of
                             armTerminateThread :
                                TerminateThread(MainThreadHandle, -1);
                             armTerminateProcess :
                                TerminateProcess(GetCurrentProcess, -1);
                             armException : raise Exception.Create('Out of time');
                           end;
                       end
                    end
                end; {else}
             end; {else of case mtsPoor..mtsRough}
           end; {case CurrentState }
        end {case WAIT_TIMEOUT}
       else
              Exit;
      end; {case}
    finally
       UnLockAccess;
    end;
  end; {While}
end;

destructor TIntruderThread.Destroy;
begin
 FreeHandlesOfMain;
 inherited Destroy;
end;

procedure TIntruderThread.LockAccess;
begin
 EnterCriticalSection(fAccessCriticalSection);
end;

procedure TIntruderThread.UnLockAccess;
begin
 LeaveCriticalSection(fAccessCriticalSection);
end;


{$IFDEF DEBUG}
initialization
 try
   AssignFile(fZ,'c:\temp\ISCdisp.txt');
   Rewrite(fz);
 finally
 end;

finalization
 try
   CloseFile(fz);
 finally
 end;

{$ENDIF}

end.
